import cv2
import numpy
from pdfplumber.page import Page

from .common import get_contours


def get_header_docket_coordinates(
    im: numpy.ndarray,
) -> tuple[int, int, int, int] | None:
    """The header represents non-title page content above the main content.

    It usually consists of three items:

    Item | Label | Example
    --:|:--|:--
    1 | Type of decision | `Resolution` or `Decision`
    2 | Page number | 1, 2, 3, etc.
    3 | Docket of the decision involved | GR. 12414, Dec. 1, 2023

    This detects Item (3) which implies that it is the in upper right quarter
    of the document:

    ```py
    x > im_w / 2 # ensures that it is on the right side of the page
    y <= im_h * 0.2 # ensures that it is on the top quarter of the page
    ```

    Item (3) is the only one above that is likely to have a second vertical line,
    hence choosing this as the the typographic bottom for the header makes sense.

    Examples:
        >>> from corpus_unpdf.src import get_page_and_img
        >>> from pathlib import Path
        >>> x = Path().cwd() / "tests" / "data" / "decision.pdf"
        >>> page, im = get_page_and_img(x, 1) # 0 marks the second page
        >>> get_header_docket_coordinates(im)
        (1813, 229, 460, 84)
        >>> page.pdf.close()

    Args:
        im (numpy.ndarray): The full page image

    Returns:
        tuple[int, int, int, int] | None: The coordinates of the docket, if found.
    """
    im_h, im_w, _ = im.shape
    for cnt in get_contours(im, (50, 50)):
        x, y, w, h = cv2.boundingRect(cnt)
        if x > im_w / 2 and y <= im_h * 0.25 and w > 200:
            return x, y, w, h
    return None


def get_header_line(im: numpy.ndarray, page: Page) -> int | float | None:
    """The header represents non-title page content above the main content.

    The terminating header line is a non-visible line that separates the
    decision's header from its main content. We'll use a typographic bottom
    of the [header][docket-coordinates] to signify this line.

    Examples:
        >>> from corpus_unpdf.src import get_page_and_img
        >>> from pathlib import Path
        >>> x = Path().cwd() / "tests" / "data" / "decision.pdf"
        >>> page, im = get_page_and_img(x, 1) # 1 marks the second page
        >>> get_header_line(im, page)
        75.12
        >>> page.pdf.close()

    Args:
        im (numpy.ndarray): The full page image
        page (Page): The pdfplumber page

    Returns:
        float | None: Y-axis point (pdfplumber point) at bottom of header
    """
    im_h, _, _ = im.shape
    if hd := get_header_docket_coordinates(im):
        _, y, _, h = hd
        header_end = (y + h) / im_h
        terminal = header_end * page.height
        return terminal
    return None


def get_page_num(page: Page, header_line: int | float) -> int | None:
    """Get the first matching digit in the header's text. This helps
    deal with decisions having blank pages.

    Examples:
        >>> import pdfplumber
        >>> from pathlib import Path
        >>> from corpus_unpdf.src.common.fetch import get_img_from_page
        >>> x = Path().cwd() / "tests" / "data" / "decision.pdf"
        >>> pdf = pdfplumber.open(x)
        >>> page = pdf.pages[1] # page 2
        >>> im = get_img_from_page(page)
        >>> header_line = get_header_line(im, page)
        >>> get_page_num(page, header_line)
        2
        >>> pdf.close()

    Args:
        page (Page): The pdfplumber page
        header_line (int | float): The value retrieved from `get_header_line()`

    Returns:
        int | None: The page number, if found
    """
    box = (0, 0, page.width, header_line)
    header = page.crop(box, relative=False, strict=True)
    texts = header.extract_text(layout=True, keep_blank_chars=True).split()
    for text in texts:
        if text.isdigit() and len(text) <= 3:
            return int(text)
    return None



from .extract import Bodyline, Footnote
from .fetch import (
    get_img_from_page,
    get_page_and_img,
    get_pages_and_imgs,
    get_reverse_pages_and_imgs,
)
from .slice import (
    PageCut,
    get_centered_coordinates,
    get_contours,
    is_match_text,
)



from .common import (
    Bodyline,
    Footnote,
    PageCut,
    get_centered_coordinates,
    get_contours,
    get_img_from_page,
    get_page_and_img,
    get_reverse_pages_and_imgs,
)
from .content_ender import get_end_page_pos
from .content_markers import (
    CourtCompositionChoices,
    DecisionCategoryChoices,
    PositionCourtComposition,
    PositionDecisionCategoryWriter,
    PositionNotice,
)
from .content_starter import get_start_page_pos
from .page_footer import get_footer_line_coordinates, get_page_end
from .page_header import get_header_line, get_page_num



import re
from typing import NamedTuple, Self

from pdfplumber.page import CroppedPage

line_break = re.compile(r"\s*\n+\s*")

paragraph_break = re.compile(r"\s{10,}(?=[A-Z])")

footnote_nums = re.compile(r"\n\s+(?P<fn>\d+)(?=\s+[A-Z])")


class Bodyline(NamedTuple):
    num: int
    line: str

    @classmethod
    def extract_lines(cls, text: str) -> list[Self]:
        """Get paragraphs using regex `\\s{10,}(?=[A-Z])`
        implying many spaces before a capital letter then
        remove new lines contained in non-paragraph lines.

        Args:
            text (str): Presumes pdfplumber.extract_text

        Returns:
            list[Self]: Bodylines of segmented text
        """
        lines = []
        for num, par in enumerate(paragraph_break.split(text), start=1):
            obj = cls(num=num, line=line_break.sub(" ", par).strip())
            lines.append(obj)
        lines.sort(key=lambda obj: obj.num)
        return lines

    @classmethod
    def from_cropped(cls, crop: CroppedPage) -> list[Self]:
        return cls.extract_lines(
            crop.extract_text(layout=True, keep_blank_chars=True)
        )


class Footnote(NamedTuple):
    fn_id: int
    note: str

    @classmethod
    def extract_notes(cls, text: str) -> list[Self]:
        """Get footnote digits using regex `\\n\\s+(?P<fn>\\d+)(?=\\s+[A-Z])`
        then for each matching span, the start span becomes the anchor
        for the balance of the text for each remaining foornote in the while
        loop. The while loop extraction must use `.pop()` where the last
        item is removed first.

        Args:
            text (str): Presumes pdfplumber.extract_text

        Returns:
            list[Self]: Footnotes separated by digits.
        """
        notes = []
        while True:
            matches = list(footnote_nums.finditer(text))
            if not matches:
                break
            note = matches.pop()  # start from the last
            footnote_num = int(note.group("fn"))
            digit_start, digit_end = note.span()
            footnote_body = text[digit_end:].strip()
            obj = cls(fn_id=footnote_num, note=footnote_body)
            notes.append(obj)
            text = text[:digit_start]
        notes.sort(key=lambda obj: obj.fn_id)
        return notes

    @classmethod
    def from_cropped(cls, crop: CroppedPage) -> list[Self]:
        return cls.extract_notes(
            crop.extract_text(layout=True, keep_blank_chars=True)
        )



from collections.abc import Iterator
from dataclasses import dataclass, field
from pathlib import Path
from typing import NamedTuple, Self

import numpy
import pdfplumber
from loguru import logger
from pdfplumber.page import CroppedPage, Page
from pdfplumber.pdf import PDF

from .src import (
    Bodyline,
    CourtCompositionChoices,
    DecisionCategoryChoices,
    Footnote,
    PageCut,
    PositionCourtComposition,
    PositionDecisionCategoryWriter,
    PositionNotice,
    get_end_page_pos,
    get_header_line,
    get_img_from_page,
    get_page_end,
    get_page_num,
    get_start_page_pos,
)


def _err(page: Page, msg: str) -> Exception:
    page.pdf.close()
    msg += f" in {page.page_number=}"
    logger.error(msg)
    raise Exception(msg)


@dataclass
class DecisionPage:
    """Metadata of a single page of the pdf file parsed via `get_decision()`

    Field | Description
    --:|:--
    `page_num` | The page number of the Decision page
    `body` | The main content above the annex, if existing
    `lines` | Segments of the `body`'s text in the given `page_num`
    `annex` | Portion of page containing the footnotes; some pages are annex-free
    `footnotes` | Each footnote item in the `annex`'s text in the given `page_num`
    """

    page_num: int
    body: CroppedPage
    annex: CroppedPage | None = None
    lines: list[Bodyline] = field(default_factory=list)
    footnotes: list[Footnote] = field(default_factory=list)

    def __post_init__(self):
        self.lines = Bodyline.from_cropped(self.body)
        if self.annex:
            self.footnotes = Footnote.from_cropped(self.annex)

    @classmethod
    def extract_proper(
        cls,
        page: Page,
        start_y: float | int | None = None,
        end_y: float | int | None = None,
    ) -> Self:
        """
        The presence of a `header_line` and a `page_endline` determine
        what to extract from a given `page`.

        The `header_line` is the imaginary line at the top of the page.
        If the `start_y` is supplied, it means that the `header_line`
        no longer needs to be calculated.

        The `page_line` is the imaginary line at the bottom of the page
        If the `end_y` is supplied, it means that the calculated `page_line`
        ought to be replaced.

        Args:
            page (Page): The pdfplumber page to evaluate
            start_y (float | int | None, optional): If present, refers to
                The y-axis point of the starter page. Defaults to None.
            end_y (float | int | None, optional): If present, refers to
                The y-axis point of the ender page. Defaults to None.

        Returns:
            Self: Page with individual components mapped out.
        """
        im = get_img_from_page(page)

        header_line = start_y or get_header_line(im, page)
        if not header_line:
            raise _err(page, "No header line")

        end_of_content, e = get_page_end(im, page)
        page_line = end_y or end_of_content

        page_num = get_page_num(page, header_line) or 0
        body = PageCut.set(page=page, y0=header_line, y1=page_line)
        annex = PageCut.set(page=page, y0=end_of_content, y1=e) if e else None
        return cls(page_num=page_num, body=body, annex=(annex))


@dataclass
class Decision:
    """Metadata of a pdf file parsed via `get_decision()`

    Field | Description
    --:|:--
    header | The top portion of the page, usually excluded from metadata
    composition | The composition of the Supreme Court that decided the case
    category | When available, whether the case is a "Decision" or a "Resolution"
    writer | When available, the writer of the case
    notice | When True, means that there is no `category` available
    pages | A list of [Decision Pages with bodies/annexes][decision-page]
    """

    composition: CourtCompositionChoices
    category: DecisionCategoryChoices | None = None
    header: CroppedPage | None = None
    writer: str | None = None
    notice: bool = False
    pages: list[DecisionPage] = field(default_factory=list)


class DecisionMeta(NamedTuple):
    start_index: int
    start_page_num: int
    start_indicator: PositionDecisionCategoryWriter | PositionNotice
    end_page_num: int
    end_page_pos: float | int

    @classmethod
    def prep(cls, path: Path):
        if not (starter := get_start_page_pos(path)):
            raise Exception("Could not detect start of content.")

        index, start_indicator = starter
        if not start_indicator:
            raise Exception("Could not detect start indicator.")

        ender = get_end_page_pos(path)
        if not ender:
            raise Exception("Could not detect end of content.")
        end_page_num, end_page_pos = ender

        return cls(
            start_index=index,
            start_page_num=index + 1,
            start_indicator=start_indicator,
            end_page_num=end_page_num,
            end_page_pos=end_page_pos,
        )

    def init(self, pdf: PDF) -> Decision:
        """Add the metadata of a Decision and extract the first page of the content
        proper which will not necessarily be page 1.

        Returns:
            Decision: A Decision instance, if all elements match.
        """
        logger.debug(f"Initialize {self=}")
        start_page = pdf.pages[self.start_index]
        if isinstance(self.start_indicator, PositionNotice):
            return Decision(
                composition=PositionCourtComposition.from_pdf(pdf).element,
                notice=True,
                pages=[
                    DecisionPage.extract_proper(
                        page=start_page,
                        start_y=self.start_indicator.position_pct_height
                        * start_page.height,
                    )
                ],
            )
        elif isinstance(self.start_indicator, PositionDecisionCategoryWriter):
            return Decision(
                composition=PositionCourtComposition.from_pdf(pdf).element,
                category=self.start_indicator.element,
                writer=self.start_indicator.writer,
                pages=[
                    DecisionPage.extract_proper(
                        page=start_page,
                        start_y=self.start_indicator.writer_pct_height
                        * start_page.height,
                    )
                ],
            )
        raise Exception("Unexpected initialization of decision.")

    def add(self, pages: list[Page]) -> Iterator[DecisionPage]:
        for nxt in pages:
            if nxt.page_number <= self.start_page_num:
                continue
            if nxt.page_number == self.end_page_num:
                logger.debug(f"Finalize {nxt.page_number=}.")
                if page_valid := DecisionPage.extract_proper(
                    page=nxt,
                    end_y=self.end_page_pos,
                ):
                    yield page_valid
                else:
                    logger.warning("Detected blank page.")
                break
            else:
                logger.debug(f"Initialize {nxt.page_number=}.")
                if page_valid := DecisionPage.extract_proper(page=nxt):
                    yield page_valid
                else:
                    logger.warning("Detected blank page.")


def get_decision(path: Path) -> Decision:
    """From a pdf file, get metadata filled Decision with pages
    cropped into bodies and annexes until the terminal page.

    Examples:
        >>> from pathlib import Path
        >>> x = Path().cwd() / "tests" / "data" / "decision.pdf"
        >>> decision = get_decision(x)
        >>> decision.category
        <DecisionCategoryChoices.RESO: 'Resolution'>
        >>> decision.composition
        <CourtCompositionChoices.DIV2: 'Second Division'>
        >>> decision.writer
        'CARPIO. J.:'
        >>> isinstance(decision.pages[0].lines[0], Bodyline)
        True
        >>> decision.pages[0].footnotes == [] # none found, # TODO: can't seem to detect
        True

    Args:
        path (Path): Path to the pdf file.

    Returns:
        Self: Instance of a Decision with pages populated
    """  # noqa: E501
    meta = DecisionMeta.prep(path)
    with pdfplumber.open(path) as pdf:
        caso = meta.init(pdf=pdf)
        content_pages = meta.add(pages=pdf.pages)
        caso.pages.extend(content_pages)
        return caso
